use std::fmt::Display;

use rgb::Rgb;
use serde::{de::Visitor, ser::{SerializeSeq, SerializeTuple}, Deserialize, Serialize};

#[derive(Debug, Clone, PartialEq, Eq)]
pub enum Variant {
    SignedByte(i8),
    Byte(u8),
    UInt(u32),
    Int(i32),
    BigUInt(u64),
    BigInt(i64),
    Boolean(bool),
    String(String),
    RGB(Rgb<u8>),
    Vec(Vec<Variant>)
}

impl Serialize for Variant {
    fn serialize<S>(&self, serializer: S) -> Result<S::Ok, S::Error>
        where
            S: serde::Serializer {
        match self {
            Variant::SignedByte(b) => serializer.serialize_i8(*b),
            Variant::Byte(b) => serializer.serialize_u8(*b),
            Variant::UInt(b) => serializer.serialize_u32(*b),
            Variant::Int(b) => serializer.serialize_i32(*b),
            Variant::BigUInt(b) => serializer.serialize_u64(*b),
            Variant::BigInt(b) => serializer.serialize_i64(*b),
            Variant::Boolean(b) => serializer.serialize_bool(*b),
            Variant::String(b) => serializer.serialize_str(&b),
            Variant::RGB(rgb) => {
                let mut seq = serializer.serialize_tuple(3).unwrap();
                seq.serialize_element(&rgb.r).unwrap();
                seq.serialize_element(&rgb.g).unwrap();
                seq.serialize_element(&rgb.b).unwrap();
                seq.end()
            },
            Variant::Vec(v) => {
                let mut seq = serializer.serialize_seq(Some(v.len())).unwrap();
                for i in v {
                    seq.serialize_element(i).unwrap();
                }
                seq.end()
            }
        }
    }
}

struct VariantVisitor {}

impl<'de> Visitor<'de> for VariantVisitor {
    type Value = Variant;

    fn expecting(&self, formatter: &mut std::fmt::Formatter) -> std::fmt::Result {
        formatter.write_str("variant")
    }

    fn visit_bool<E>(self, v: bool) -> Result<Self::Value, E>
        where
            E: serde::de::Error, {
        Ok(Variant::Boolean(v))
    }

    fn visit_str<E>(self, v: &str) -> Result<Self::Value, E>
        where
            E: serde::de::Error, {
        Ok(Variant::String(v.into()))
    }

    fn visit_string<E>(self, v: String) -> Result<Self::Value, E>
        where
            E: serde::de::Error, {
        Ok(Variant::String(v))
    }
}

impl<'de> Deserialize<'de> for Variant {
    fn deserialize<D>(deserializer: D) -> Result<Self, D::Error>
        where
            D: serde::Deserializer<'de> {
        deserializer.deserialize_any(VariantVisitor {})
    }
}

macro_rules! impl_variant_type {
    ($type:ty, $var_type:tt) => {
        impl From<$type> for Variant {
            fn from(value: $type) -> Self {
                Variant::$var_type(value)
            }
        }

        impl Into<$type> for Variant {
            fn into(self) -> $type {
                match self {
                    Variant::$var_type(value) => value,
                    _ => panic!(concat!("Expected Variant::", stringify!($var_type), "but got {:?}"), self)
                }
            }
        }

        impl Into<Vec<$type>> for Variant {
            fn into(self) -> Vec<$type> {
                match self {
                    Variant::Vec(v) => {
                        v.into_iter().map(|f| {
                            f.into()
                        }).collect()
                    },
                    _ => panic!(concat!("Expected Variant::Vec(", stringify!($var_type), "but got {:?}"), self)
                }
            }
        }
    };
}

impl_variant_type!(u8, Byte);
impl_variant_type!(i8, SignedByte);
impl_variant_type!(u32, UInt);
impl_variant_type!(i32, Int);
impl_variant_type!(i64, BigInt);
impl_variant_type!(u64, BigUInt);
impl_variant_type!(bool, Boolean);
impl_variant_type!(String, String);
impl_variant_type!(Rgb<u8>, RGB);
impl_variant_type!(Vec<Variant>, Vec);

impl<'a> From<&'a str> for Variant {
    fn from(value: &'a str) -> Self {
        Variant::String(value.to_owned())
    }
}

#[derive(Debug, Clone, PartialEq, Eq, Serialize)]
pub struct Property {
    key: PropertyKey,
    value: Variant
}

impl Property {
    pub const fn new(key: PropertyKey, value: Variant) -> Self {
        Property {
            key,
            value,
        }
    }
}

#[derive(Debug, Clone, Copy, Eq, PartialEq, Serialize, Deserialize)]
pub struct NamespaceKey(pub &'static str);

#[derive(Debug, Clone, Copy, Eq, PartialEq, Serialize, Deserialize)]
pub struct PropertyKey(pub &'static str);

#[derive(Debug, Clone, Copy, Eq, PartialEq, Serialize)]
pub struct PropertyID {
    pub namespace: NamespaceKey,
    pub key: PropertyKey
}

impl<'a> Display for PropertyID {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        write!(f, "{}::{}", self.namespace.0, self.key.0)
    }
}

pub trait Namespace {
    fn nskey() -> NamespaceKey;
    fn properties() -> Vec<Property>;
}

#[macro_export]
macro_rules! property_namespace {
    ($name:ident, $($prop_name:ident => $default_value:expr),*) => {
        pub struct $name {}

        impl $name {
            #[allow(non_upper_case_globals)]
            pub const Namespace: $crate::properties::NamespaceKey = $crate::properties::NamespaceKey(stringify!($name));
            $(
                #[allow(non_upper_case_globals)]
                pub const $prop_name: $crate::properties::PropertyID = $crate::properties::PropertyID {
                    namespace: Self::Namespace,
                    key: $crate::properties::PropertyKey(stringify!($prop_name))
                };
            )*
        }
        
        impl crate::properties::Namespace for $name {
            fn nskey() -> $crate::properties::NamespaceKey {
                Self::Namespace
            }

            fn properties() -> Vec<$crate::properties::Property> {
                vec![
                    $($crate::properties::Property::new(Self::$prop_name.key, $crate::properties::Variant::from($default_value)),)*
                ]
            }
        }
    };
}

#[derive(Debug, Clone)]
pub struct NamespaceProps {
    props: Vec<Property>,
    key: NamespaceKey
}

impl NamespaceProps {
    fn new<NS: Namespace>() -> Self {
        NamespaceProps {
            props: NS::properties(),
            key: NS::nskey()
        }
    }

    fn get_key(&self, key: &PropertyKey) -> Option<&Property> {
        for next in self.props.iter() {
            if next.key == *key {
                return Some(next);
            }
        }

        log::warn!("Unknown key {:?}", key);

        return None;
    }

    fn get_key_mut(&mut self, key: &PropertyKey) -> Option<&mut Property> {
        for next in self.props.iter_mut() {
            if next.key == *key {
                return Some(next);
            }
        }

        log::warn!("Unknown key {:?}", key);

        return None;
    }

}

#[derive(Debug, Clone)]
pub struct Properties {
    contents: Vec<NamespaceProps>
}

impl Display for Properties {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        for ns in self.contents.iter() {
            write!(f, "{}\n", ns.key.0).unwrap();
            for prop in ns.props.iter() {
                write!(f, "\t{} = {:?}\n", prop.key.0, prop.value).unwrap();
            }
        }
        Result::Ok(())
    }
}

impl Properties {
    pub fn new() -> Self {
        Properties {
            contents: Vec::new(),
        }
    }

    pub fn add_namespace<T: Namespace>(&mut self) {
        self.contents.push(NamespaceProps::new::<T>());
    }

    fn get_namespace(&self, ns: &NamespaceKey) -> Option<&NamespaceProps> {
        for nsprops in self.contents.iter() {
            if nsprops.key == *ns {
                return Some(nsprops)
            }
        }

        log::warn!("Unknown namespace {:?}", ns);

        None
    }

    fn get_namespace_mut(&mut self, ns: &NamespaceKey) -> Option<&mut NamespaceProps> {
        for nsprops in self.contents.iter_mut() {
            if nsprops.key == *ns {
                return Some(nsprops)
            }
        }

        log::warn!("Unknown namespace {:?}", ns);

        None
    }

    pub fn get<T: Into<PropertyID>>(&self, key: T) -> Option<Variant> {
        let as_id = key.into();
        match self.get_namespace(&as_id.namespace).unwrap().get_key(&as_id.key) {
            None => None,
            Some(v) => Some(v.value.clone())
        }
    }

    pub fn set<V, T: Into<PropertyID>>(&mut self, key: T, value: V) -> bool where Variant: From<V> {
        let as_id = key.into();
        let as_variant: Variant = value.into();
        match self.get_namespace_mut(&as_id.namespace).unwrap().get_key_mut(&as_id.key) {
            None => (),
            Some(found_key) => {
                if found_key.value != as_variant {
                    found_key.value = as_variant;
                    return true
                }
            }
        }

        return false
    }

    pub fn all_props(&self) -> Vec<PropertyID> {
        self.contents.iter().map(|f| {
            f.props.iter().map(|k| {
                PropertyID { namespace: f.key, key: k.key }
            })
        }).flatten().collect()
    }
}