use core::fmt;

use crate::{events::{Event, EventBus}, properties::{Properties, PropertyID, Variant}, EventSource, Evented};

pub struct Environment<'a> {
    pub props: &'a mut Properties,
    pub bus: &'a mut EventBus
}

impl Environment {
    
}

impl<'a> Environment<'a> {
    pub fn get_property<K: Into<PropertyID>, V>(&mut self, key: K) -> Option<V> where Variant: Into<V> {
        self.props.get(key.into()).and_then(|f| { Some(f.into()) })
    }

    pub fn set_property<K: Into<PropertyID>, V: Into<Variant>>(&mut self, key: K, value: V) {
        self.bus.push(Event::new_property_change(key.into(), value.into()));
    }
}

pub trait Task: Send {
    fn on_ready(&mut self, env: &mut Environment) {}
    fn on_tick(&mut self, env: &mut Environment) {}
    fn on_property_change(&mut self, key: PropertyID, value: &Variant, env: &mut Environment) {}
    
    fn start(&mut self, env: &mut Environment) {}
    fn stop(&mut self, env: &mut Environment) {}
    fn name(&self) -> &'static str {
        core::any::type_name::<Self>()
    }
}

#[derive(Debug, Clone)]
enum ScheduledState {
    Stopped,
    Start,
    Running,
    Stop
}

struct ScheduledTask {
    state: ScheduledState,
    task: Box<dyn Task>,
}

impl core::fmt::Debug for ScheduledTask {
    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
        f.debug_struct("ScheduledTask")
            .field("task", &self.task.name())
            .field("state", &self.state)
            .finish()
    }
}

impl ScheduledTask {
    fn new(task: Box<dyn Task>) -> Self {
        ScheduledTask {
            state: ScheduledState::Stopped,
            task: task,
        }
    }

    fn start(&mut self) {
        match self.state {
            ScheduledState::Stopped => self.state = ScheduledState::Start,
            ScheduledState::Stop => self.state = ScheduledState::Running,
            _ => ()
        }
    }

    fn stop(&mut self) {
        match self.state {
            ScheduledState::Running => self.state = ScheduledState::Stop,
            ScheduledState::Start => self.state = ScheduledState::Stopped,
            _ => ()
        }
    }

    fn on_event(&mut self, event: &Event, env: &mut Environment) {
        match self.state {
            ScheduledState::Start => {
                log::info!("Starting task {}", self.task.name());
                self.task.start(env);
                self.state = ScheduledState::Running
            },
            ScheduledState::Stop => {
                log::info!("Stopping task {}", self.task.name());
                self.task.stop(env);
                self.state = ScheduledState::Stopped
            },
            _ => ()
        };

        match self.state {
            ScheduledState::Running => {
                match event {
                    Event::Tick => self.task.on_tick(env),
                    Event::ReadyToRock => self.task.on_ready(env),
                    Event::PropertyChange(key, value) => self.task.on_property_change(key.clone(), value, env),
                    _ => ()
                }
            },
            _ => ()
        }
    }
}

#[derive(Debug)]
pub struct FixedSizeScheduler<const TASK_COUNT: usize> {
    tasks: [ScheduledTask; TASK_COUNT],
}

impl<const TASK_COUNT: usize> FixedSizeScheduler<TASK_COUNT> {
pub fn new(tasks: [Box<dyn Task>; TASK_COUNT]) -> Self {
        let mut scheduled: [ScheduledTask; TASK_COUNT] = unsafe { std::mem::MaybeUninit::zeroed().assume_init() };
        let mut idx = 0;
        for task in tasks {
            log::info!("Scheduling task {}", task.name());
            let slot = &mut scheduled[idx];
            unsafe { std::ptr::write(slot, ScheduledTask::new(task)) };
            idx += 1;
        }
        FixedSizeScheduler {
            tasks: scheduled
        }
    }

    fn find_task(&mut self, name: &str) -> Option<&mut ScheduledTask> {
        for slot in &mut self.tasks {
            if slot.task.name() == name {
                return Some(slot);
            }
        }

        None
    }

    fn on_event(&mut self, event: &Event, env: &mut Environment) {
        match event {
            Event::StartThing(task_name) => {
                if let Some(slot) = self.find_task(task_name) {
                    log::info!("Queuing start for {}", task_name);
                    slot.start();
                }
            },
            Event::StopThing(task_name) => {
                if let Some(slot) = self.find_task(task_name) {
                    log::info!("Queuing stop for {}", task_name);
                    slot.stop();
                }
            },
            _ => {
                for slot in &mut self.tasks {
                    slot.on_event(event, env);
                }
            }
        }
    }
}